#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.errors as common_errors
import octobot_commons.databases as databases

import octobot_backtesting.data as data
import octobot_backtesting.enums as enums
import octobot_backtesting.errors as errors
import octobot_backtesting.importers as importers


class ExchangeDataImporter(importers.DataImporter):
    def __init__(self, config, file_path):
        super().__init__(config, file_path)

        self.exchange_name = None
        self.symbols = []
        self.time_frames = []
        self.available_data_types = []
        self.has_all_time_frames_candles_history = False

    async def initialize(self) -> None:
        self.load_database()
        await self.database.initialize()

        # load description
        description = await data.get_database_description(self.database)
        self.exchange_name = description[enums.DataFormatKeys.EXCHANGE.value]
        self.symbols = description[enums.DataFormatKeys.SYMBOLS.value]
        self.time_frames = description[enums.DataFormatKeys.TIME_FRAMES.value]
        self.has_all_time_frames_candles_history = bool(description.get(enums.DataFormatKeys.START_TIMESTAMP.value))
        await self._init_available_data_types()

        self.logger.info(f"Loaded {self.exchange_name} data file with "
                         f"{', '.join(self.symbols)} on {', '.join([tf.value for tf in self.time_frames])}")

    async def start(self) -> None:
        pass

    async def get_data_timestamp_interval(self, time_frame=None):
        minimum_timestamp: float = 0.0
        maximum_timestamp: float = 0.0

        min_ohlcv_timestamp: float = 0.0
        max_ohlcv_timestamp: float = 0.0

        for table in [enums.ExchangeDataTables.KLINE, enums.ExchangeDataTables.ORDER_BOOK,
                      enums.ExchangeDataTables.RECENT_TRADES, enums.ExchangeDataTables.TICKER]:
            if table in self.available_data_types:
                try:
                    min_timestamp = (await self.database.select_min(table,
                                                                    [databases.SQLiteDatabase.TIMESTAMP_COLUMN]))[0][0]
                    if not minimum_timestamp or minimum_timestamp > min_timestamp:
                        minimum_timestamp = min_timestamp

                    max_timestamp = (await self.database.select_max(table,
                                                                    [databases.SQLiteDatabase.TIMESTAMP_COLUMN]))[0][0]
                    if not maximum_timestamp or maximum_timestamp < max_timestamp:
                        maximum_timestamp = max_timestamp
                except (IndexError, common_errors.DatabaseNotFoundError):
                    pass

        # OHLCV timestamps
        try:
            ohlcv_kwargs = {"time_frame": time_frame} if time_frame else {}
            ohlcv_min_timestamps = (await self.database.select_min(enums.ExchangeDataTables.OHLCV,
                                                                   [databases.SQLiteDatabase.TIMESTAMP_COLUMN],
                                                                   [common_constants.CONFIG_TIME_FRAME],
                                                                   group_by=common_constants.CONFIG_TIME_FRAME,
                                                                   **ohlcv_kwargs
                                                                   ))

            if ohlcv_min_timestamps:
                # if the required time frame is not included in this database, ohlcv_min_timestamps is empty: ignore it
                min_ohlcv_timestamp = max(ohlcv_min_timestamps)[0]
                max_ohlcv_timestamp = (await self.database.select_max(enums.ExchangeDataTables.OHLCV,
                                                                      [databases.SQLiteDatabase.TIMESTAMP_COLUMN],
                                                                      **ohlcv_kwargs))[0][0]
            elif time_frame:
                raise errors.MissingTimeFrame(f"Missing time frame in data file: {time_frame}")

        except (IndexError, common_errors.DatabaseNotFoundError):
            pass

        if minimum_timestamp > 0 and maximum_timestamp > 0:
            return max(minimum_timestamp, min_ohlcv_timestamp), max(maximum_timestamp, max_ohlcv_timestamp)
        return min_ohlcv_timestamp, max_ohlcv_timestamp

    async def _init_available_data_types(self):
        self.available_data_types = [table for table in enums.ExchangeDataTables
                                     if await self.database.check_table_exists(table)
                                     and await self.database.check_table_not_empty(table)]

    async def _get_from_db(
            self, exchange_name, symbol, table,
            time_frame=None,
            limit=databases.SQLiteDatabase.DEFAULT_SIZE,
            timestamps=None,
            operations=None
    ):
        kwargs = {} if time_frame is None else {"time_frame": time_frame.value}

        if timestamps:
            return await self.database.select_from_timestamp(
                table, size=limit,
                exchange_name=exchange_name, symbol=symbol,
                timestamps=timestamps,
                operations=operations,
                **kwargs
            )
        return await self.database.select(
            table, size=limit,
            exchange_name=exchange_name, symbol=symbol,
            **kwargs
        )

    async def get_ohlcv(self, exchange_name=None, symbol=None,
                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                        limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                        timestamps=None,
                        operations=None):
        return importers.import_ohlcvs(await self._get_from_db(
            exchange_name, symbol, enums.ExchangeDataTables.OHLCV,
            time_frame=time_frame,
            limit=limit,
            timestamps=timestamps,
            operations=operations
        ))

    async def get_ohlcv_from_timestamps(self, exchange_name=None, symbol=None,
                                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                                        limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                        inferior_timestamp=-1, superior_timestamp=-1) -> list:
        """
        Reads OHLCV history from database and populates a local ChronologicalReadDatabaseCache.
        Warning: can't read data from before last given inferior_timestamp unless associated cache is reset
        """
        return await self._get_from_cache(exchange_name, symbol, time_frame, enums.ExchangeDataTables.OHLCV,
                                          inferior_timestamp, superior_timestamp, self.get_ohlcv, limit)

    async def get_ticker(self, exchange_name=None, symbol=None,
                         limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                         timestamps=None,
                         operations=None):
        return importers.import_tickers(await self._get_from_db(
            exchange_name, symbol, enums.ExchangeDataTables.TICKER,
            limit=limit,
            timestamps=timestamps,
            operations=operations
        ))

    async def get_ticker_from_timestamps(self, exchange_name=None, symbol=None,
                                         limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                         inferior_timestamp=-1, superior_timestamp=-1):
        """
        Reads ticker history from database and populates a local ChronologicalReadDatabaseCache.
        Warning: can't read data from before last given inferior_timestamp unless associated cache is reset
        """
        return await self._get_from_cache(exchange_name, symbol, None, enums.ExchangeDataTables.TICKER,
                                          inferior_timestamp, superior_timestamp, self.get_ticker, limit)

    async def get_order_book(self, exchange_name=None, symbol=None,
                             limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                             timestamps=None,
                             operations=None):
        return importers.import_order_books(await self._get_from_db(
            exchange_name, symbol, enums.ExchangeDataTables.ORDER_BOOK,
            limit=limit,
            timestamps=timestamps,
            operations=operations
        ))

    async def get_order_book_from_timestamps(self, exchange_name=None, symbol=None,
                                             limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                             inferior_timestamp=-1, superior_timestamp=-1):
        """
        Reads order book history from database and populates a local ChronologicalReadDatabaseCache.
        Warning: can't read data from before last given inferior_timestamp unless associated cache is reset
        """
        return await self._get_from_cache(exchange_name, symbol, None, enums.ExchangeDataTables.ORDER_BOOK,
                                          inferior_timestamp, superior_timestamp, self.get_order_book, limit)

    async def get_recent_trades(self, exchange_name=None, symbol=None,
                                limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                timestamps=None,
                                operations=None):
        return importers.import_recent_trades(await self._get_from_db(
            exchange_name, symbol, enums.ExchangeDataTables.RECENT_TRADES,
            limit=limit,
            timestamps=timestamps,
            operations=operations
        ))

    async def get_recent_trades_from_timestamps(self, exchange_name=None, symbol=None,
                                                limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                                inferior_timestamp=-1, superior_timestamp=-1):
        """
        Reads recent trades history from database and populates a local ChronologicalReadDatabaseCache.
        Warning: can't read data from before last given inferior_timestamp unless associated cache is reset
        """
        return await self._get_from_cache(exchange_name, symbol, None, enums.ExchangeDataTables.RECENT_TRADES,
                                          inferior_timestamp, superior_timestamp, self.get_recent_trades, limit)

    async def get_kline(self, exchange_name=None, symbol=None,
                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                        limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                        timestamps=None,
                        operations=None):
        return importers.import_klines(await self._get_from_db(
            exchange_name, symbol, enums.ExchangeDataTables.KLINE,
            time_frame=time_frame,
            limit=limit,
            timestamps=timestamps,
            operations=operations
        ))

    async def get_kline_from_timestamps(self, exchange_name=None, symbol=None,
                                        time_frame=common_enums.TimeFrames.ONE_HOUR,
                                        limit=databases.SQLiteDatabase.DEFAULT_SIZE,
                                        inferior_timestamp=-1, superior_timestamp=-1):
        """
        Reads kline history from database and populates a local ChronologicalReadDatabaseCache.
        Warning: can't read data from before last given inferior_timestamp unless associated cache is reset
        """
        return await self._get_from_cache(exchange_name, symbol, time_frame, enums.ExchangeDataTables.KLINE,
                                          inferior_timestamp, superior_timestamp, self.get_kline, limit)

    async def _get_from_cache(self, exchange_name, symbol, time_frame, data_type,
                              inferior_timestamp, superior_timestamp, set_cache_method, limit):
        if not self.chronological_cache.has((exchange_name, symbol, time_frame, data_type)):
            # ignore superior timestamp to select everything starting from inferior_timestamp and cache it
            select_superior_timestamp = -1
            timestamps, operations = importers.get_operations_from_timestamps(
                select_superior_timestamp,
                inferior_timestamp
            )
            # initializer without time_frame args are not expecting the time_frame argument, remove it
            # ignore the limit param as it might reduce the available cache and give false later select results
            init_cache_method_args = \
                (exchange_name, symbol, databases.SQLiteDatabase.DEFAULT_SIZE, timestamps, operations) \
                if time_frame is None \
                else (exchange_name, symbol, time_frame, databases.SQLiteDatabase.DEFAULT_SIZE, timestamps, operations)
            self.chronological_cache.set(
                await set_cache_method(*init_cache_method_args),
                0,
                (exchange_name, symbol, time_frame, data_type)
            )
        return self.chronological_cache.get(inferior_timestamp, superior_timestamp,
                                            (exchange_name, symbol, time_frame, data_type))



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json

import octobot_commons.enums as commons_enums


def get_operations_from_timestamps(superior_timestamp, inferior_timestamp):
    operations: list = []
    timestamps: list = []
    if superior_timestamp != -1:
        timestamps.append(str(superior_timestamp))
        operations.append(commons_enums.DataBaseOperations.INF_EQUALS.value)
    if inferior_timestamp != -1:
        timestamps.append(str(inferior_timestamp))
        operations.append(commons_enums.DataBaseOperations.SUP_EQUALS.value)

    return timestamps, operations


def import_ohlcvs(ohlcvs):
    for i, val in enumerate(ohlcvs):
        ohlcvs[i] = list(val)
        ohlcvs[i][-1] = json.loads(ohlcvs[i][-1])
    return ohlcvs


def import_tickers(tickers):
    for ticker in tickers:
        ticker[-1] = json.loads(ticker[-1])
    return tickers


def import_order_books(order_books):
    for order_book in order_books:
        order_book[-1] = json.loads(order_book[-1])
        order_book[-2] = json.loads(order_book[-2])
    return order_books


def import_recent_trades(recent_trades):
    for recent_trade in recent_trades:
        recent_trade[-1] = json.loads(recent_trade[-1])
    return recent_trades


def import_klines(klines):
    for kline in klines:
        kline[-1] = json.loads(kline[-1])
    return klines



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import copy

import async_channel.channels as channels
import async_channel.enums as channel_enums

import octobot_commons.channels_name as channels_name
import octobot_commons.list_util as list_util
import octobot_commons.logging as logging
import octobot_commons.asyncio_tools as asyncio_tools


class ChannelsManager:
    DEFAULT_REFRESH_TIMEOUT = 15

    def __init__(self, exchange_ids, matrix_id, time_chan_name, refresh_timeout=DEFAULT_REFRESH_TIMEOUT):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.exchange_ids = exchange_ids
        self.matrix_id = matrix_id
        self.time_chan_name = time_chan_name
        self.refresh_timeout = refresh_timeout
        self.producers = []
        self.initial_producers = []
        self.iteration_task = None
        self.should_stop = False
        self.producers_by_priority_levels = {}

    async def initialize(self) -> None:
        """
        Initialize Backtesting channels manager
        """
        self.logger.debug("Initializing producers...")
        try:
            self.initial_producers = list_util.flatten_list(
                _get_backtesting_producers(self.time_chan_name) +
                self._get_trading_producers() +
                self._get_evaluator_producers()
            )
            self.producers = copy.copy(self.initial_producers)

            self.producers_by_priority_levels = {
                priority_level.value: self.producers
                for priority_level in channel_enums.ChannelConsumerPriorityLevels
            }

            # Initialize all producers by calling producer.start()
            for producer in list_util.flatten_list(self._get_trading_producers() + self._get_evaluator_producers()):
                await producer.start()
        except Exception as exception:
            self.logger.exception(exception, True, f"Error when initializing backtesting: {exception}")
            raise

    def clear_empty_channels_producers(self):
        self.producers = [
            producer
            for producer in self.initial_producers
            if producer.channel.get_consumers()
        ]

    def update_producers_by_priority_levels(self):
        self.producers_by_priority_levels = {
            priority_level.value: _get_producers_with_priority_level_consumers(self.producers, priority_level.value)
            for priority_level in channel_enums.ChannelConsumerPriorityLevels
            if _check_producers_has_priority_consumers(self.producers, priority_level.value)
        }

    async def handle_new_iteration(self, current_timestamp) -> None:
        for level_key, producers in self.producers_by_priority_levels.items():
            try:
                if _check_producers_consumers_emptiness(producers, level_key):
                    # avoid creating tasks when not necessary
                    continue
                self.iteration_task = self.refresh_priority_level(producers, level_key, True)
                await self.iteration_task
                # trigger waiting events
                await asyncio_tools.wait_asyncio_next_cycle()
                # massive slow down
                # self.iteration_task = await asyncio.wait_for(self.refresh_priority_level(level_key.value, True),
                #                                              timeout=self.refresh_timeout)
            except asyncio.TimeoutError:
                self.logger.error(f"Refreshing priority level {level_key} has been timed out at timestamp "
                                  f"{current_timestamp}.")

    async def refresh_priority_level(self, producers, priority_level: int, join_consumers: bool) -> None:
        while not self.should_stop:
            for producer in producers:
                await producer.synchronized_perform_consumers_queue(priority_level, join_consumers, self.refresh_timeout)
            if _check_producers_consumers_emptiness(self.producers, priority_level):
                break

    def stop(self):
        self.should_stop = True

    def flush(self):
        self.producers = []
        self.initial_producers = []
        self.producers_by_priority_levels = {}
        self.iteration_task = None

    def _get_trading_producers(self):
        import octobot_trading.exchange_channel as exchange_channel
        return [
            _get_channel_producers(exchange_channel.get_chan(channel_name.value, exchange_id))
            for exchange_id in self.exchange_ids
            for channel_name in channels_name.OctoBotTradingChannelsName
        ]

    def _get_evaluator_producers(self):
        import octobot_evaluators.evaluators.channel as evaluators_channel
        return [
            _get_channel_producers(evaluators_channel.get_chan(channel_name.value, self.matrix_id))
            for channel_name in channels_name.OctoBotEvaluatorsChannelsName
        ]


def _get_channel_producers(channel):
    if channel.producers:
        return channel.producers
    return [channel.get_internal_producer()]


def _get_backtesting_producers(time_chan_name):
    return [
        _get_channel_producers(channels.get_chan(channel_name))
        for channel_name in [time_chan_name]
    ]


def _check_producers_consumers_emptiness(producers, priority_level):
    for producer in producers:
        try:
            if not producer.is_consumers_queue_empty(priority_level):
                return False
        except AttributeError:
            if producer.channel is None:
                # channel has been cleared, there is nothing to do
                return True
            else:
                # unexpected, propagate
                raise
    return True


def _check_producers_has_priority_consumers(producers, priority_level):
    for producer in producers:
        if producer.channel.get_prioritized_consumers(priority_level):
            return True
    return False


def _get_producers_with_priority_level_consumers(producers, priority_level):
    return [
        producer
        for producer in producers
        if producer.channel.get_prioritized_consumers(priority_level)
    ]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_backtesting.importers.exchanges import exchange_importer
from octobot_backtesting.importers.exchanges import util

from octobot_backtesting.importers.exchanges.exchange_importer import (
    ExchangeDataImporter,
)

from octobot_backtesting.importers.exchanges.util import (
    get_operations_from_timestamps,
    import_ohlcvs,
    import_tickers,
    import_order_books,
    import_recent_trades,
    import_klines,
)

__all__ = [
    "ExchangeDataImporter",
    "get_operations_from_timestamps",
    "import_ohlcvs",
    "import_tickers",
    "import_order_books",
    "import_recent_trades",
    "import_klines",
]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_backtesting.enums as backtesting_enums
import octobot_backtesting.importers as importers
import octobot_backtesting.util as util

import octobot_commons.errors as commons_errors
import octobot_commons.databases as databases


async def create_importer(config, backtesting_file, default_importer=None):
    return await util.create_importer_from_backtesting_file_name(config, backtesting_file,
                                                                 default_importer=default_importer)


def get_available_data_types(importer) -> list:
    return importer.available_data_types


def get_data_file(importer) -> str:
    return importer.file_path


def get_data_file_from_importers(available_importers, symbol, time_frame) -> str:
    for importer in available_importers:
        if symbol in get_available_symbols(importer) and \
                time_frame in get_available_time_frames(importer):
            return get_data_file_path(importer)
    return None


def get_data_file_path(importer) -> str:
    return importer.adapt_file_path_if_necessary()


def get_available_time_frames(exchange_importer) -> list:
    return exchange_importer.time_frames


def get_available_symbols(exchange_importer) -> list:
    return [symbol for symbol in exchange_importer.symbols]


async def get_data_timestamp_interval(exchange_importer, time_frame=None) -> (float, float):
    time_frame_value = time_frame.value if time_frame is not None else None
    return await exchange_importer.get_data_timestamp_interval(time_frame=time_frame_value)


async def get_all_ohlcvs(database_path, exchange_name, symbol, time_frame,
                         inferior_timestamp=-1, superior_timestamp=-1) -> list:
    timestamps, operations = importers.get_operations_from_timestamps(superior_timestamp, inferior_timestamp)
    try:
        async with databases.new_sqlite_database(database_path) as database:
            candles = await database.select_from_timestamp(backtesting_enums.ExchangeDataTables.OHLCV,
                                                           exchange_name=exchange_name, symbol=symbol,
                                                           time_frame=time_frame.value,
                                                           timestamps=timestamps,
                                                           operations=operations)
        return [candle_with_metadata[-1]
                for candle_with_metadata in sorted(importers.import_ohlcvs(candles), key=lambda x: x[0])]
    except commons_errors.DatabaseNotFoundError:
        return []


async def stop_importer(importer) -> None:
    await importer.stop()



import octobot_backtesting.util as util
import octobot_commons.enums as commons_enums


class BacktestData:
    def __init__(self, data_files, config, tentacles_config, use_accurate_price_time_frame):
        self.data_files = data_files
        self.config = config
        self.tentacles_config = tentacles_config
        self.importers_by_data_file = None
        self.preloaded_candle_managers = {}
        self.use_cached_markets: bool = False
        self.default_importer = None
        self.use_accurate_price_time_frame: bool = use_accurate_price_time_frame

    async def initialize(self):
        self.importers_by_data_file = {
            data_file: await util.create_importer_from_backtesting_file_name(self.config, data_file,
                                                                             default_importer=self.default_importer)
            for data_file in self.data_files
        }

    async def get_preloaded_candles_manager(self, exchange, symbol, time_frame, start_timestamp, end_timestamp):
        key = self._get_key(exchange, symbol, time_frame, start_timestamp, end_timestamp)
        try:
            return self.preloaded_candle_managers[key]
        except KeyError:
            try:
                import octobot_trading.api as trading_api
                preloaded_candles = await self._get_all_candles(exchange, symbol, time_frame,
                                                                start_timestamp, end_timestamp)
                self.preloaded_candle_managers[key] = await trading_api.create_preloaded_candles_manager(
                    preloaded_candles
                )
                return self.preloaded_candle_managers[key]
            except ImportError:
                self.preloaded_candle_managers[key] = None
        return self.preloaded_candle_managers[key]

    async def _get_all_candles(self, exchange, symbol, time_frame, start_timestamp, end_timestamp):
        # manually filter max candles not to change chronological cache behavior
        all_candles = await self._get_importer(exchange, symbol).get_ohlcv_from_timestamps(
                exchange_name=exchange,
                symbol=symbol,
                time_frame=time_frame,
                inferior_timestamp=start_timestamp,
            )
        return [
            candle[-1]
            for candle in all_candles
            if candle[-1][commons_enums.PriceIndexes.IND_PRICE_TIME.value] <= end_timestamp
        ]

    def _get_importer(self, exchange, symbol):
        for importer in self.importers_by_data_file.values():
            if exchange in importer.exchange_name == exchange and symbol in importer.symbols:
                return importer
        raise KeyError("Importer not found")

    def reset_cached_indexes(self):
        for importer in self.importers_by_data_file.values():
            importer.reset_cache()

    async def stop(self):
        for importer in self.importers_by_data_file.values():
            await importer.stop()

    def _get_key(self, *identifiers):
        return "-".join(str(i) for i in identifiers)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import async_channel.channels as channels
import async_channel.util as channel_util

import octobot_commons.logging as logging
import octobot_commons.tentacles_management as tentacles_management

import octobot_backtesting.util as backtesting_util
import octobot_backtesting.time as backtesting_time


class Backtesting:
    def __init__(self, config, exchange_ids, matrix_id, backtesting_files,
                 importers_by_data_file=None, backtest_data=None, bot_id=None):
        self.config = config
        self.backtesting_files = backtesting_files
        self.importers_by_data_file = importers_by_data_file or {}
        self.logger = logging.get_logger(self.__class__.__name__)

        self.exchange_ids = exchange_ids
        self.matrix_id = matrix_id
        self.bot_id = bot_id or ""

        self.importers = []
        self.backtest_data = backtest_data
        self.time_manager = None
        self.time_updater = None
        self.time_channel = None

    async def initialize(self):
        time_chan_name = self.get_time_chan_name()  # not in try to be able to raise on error
        try:
            self.time_manager = backtesting_time.TimeManager(config=self.config)
            self.time_manager.initialize()

            self.time_channel = await channel_util.create_channel_instance(
                backtesting_time.TimeChannel,
                channels.set_chan,
                is_synchronized=True,
                channel_name=time_chan_name
            )

            self.time_updater = backtesting_time.TimeUpdater(
                channels.get_chan(self.get_time_chan_name()),
                self
            )
        except Exception as e:
            self.logger.exception(e, True, f"Error when initializing backtesting : {e}.")

    def use_accurate_price_time_frame(self) -> bool:
        for importer in self.importers:
            if not importer.has_all_time_frames_candles_history:
                # has_all_time_frames_candles_history is necessary for accurate price time frame
                return False
        if self.backtest_data:
            return self.backtest_data.use_accurate_price_time_frame
        return True

    def get_time_chan_name(self):
        return backtesting_time.TimeChannel.get_name(self.bot_id)

    async def stop(self):
        await self.delete_time_channel()

    async def delete_time_channel(self):
        await self.time_channel.stop()
        for consumer in self.time_channel.consumers:
            await self.time_channel.remove_consumer(consumer)
        self.time_channel.flush()
        channels.del_chan(self.get_time_chan_name())

    async def start_time_updater(self):
        await self.time_updater.run()

    async def _create_importer(self, backtesting_file):
        return await backtesting_util.create_importer_from_backtesting_file_name(
            self.config,
            backtesting_file,
            backtesting_util.get_default_importer()
        )

    async def create_importers(self):
        try:
            self.importers = []
            for backtesting_file in self.backtesting_files:
                if self.importers_by_data_file is not None and backtesting_file in self.importers_by_data_file:
                    self.importers.append(self.importers_by_data_file[backtesting_file])
                else:
                    self.importers.append(await self._create_importer(backtesting_file))
        except TypeError:
            pass

    async def handle_time_update(self, timestamp):
        if self.time_manager:
            self.time_manager.set_current_timestamp(timestamp)

    def get_importers(self, importer_parent_class=None) -> list:
        return [importer
                for importer in self.importers
                if tentacles_management.default_parents_inspection(importer.__class__, importer_parent_class)] \
            if importer_parent_class is not None else self.importers

    def get_progress(self):
        if self._has_nothing_to_do():
            return 0
        return 1 - (self.time_manager.get_remaining_iteration() / self.time_manager.get_total_iteration())

    def is_in_progress(self):
        if self._has_nothing_to_do():
            return False
        else:
            return self.time_manager.get_remaining_iteration() > 0

    def _has_nothing_to_do(self):
        return not self.time_manager or self.time_manager.get_total_iteration() == 0

    def has_finished(self):
        return self.time_updater and self.time_updater.finished_event.is_set()
